#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/xpm.h>
#include "gnu.xpm"

enum {TEST_BLIT, TEST_FILL};

void redraw(void);
Window create_window(int xsz, int ysz, const char *title);
int handle_event(XEvent *ev);
void post_redisplay(void);

Display *dpy;
Window win;
Pixmap pixmap;
Atom wm_del_win, wm_prot;
GC gc;

int win_width = 800;
int win_height = 600;
int pmap_width, pmap_height;
int frames;

int test;

#define SEC(u)	((float)(u) / 1000000.0f)

int main(int argc, char **argv)
{
	int i, tmp;
	Window root_win;
	XEvent event;
	struct timeval tv0, tv;
	float sec;

	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-' && argv[i][2] == 0) {
			switch(argv[i][1]) {
			case 'f':
				test = TEST_FILL;
				break;

			case 'b':
				test = TEST_BLIT;
				break;

			case 's':
				{
					char *xptr;
					if(!isdigit(argv[++i][0]) ||
							!(xptr = strchr(argv[i], 'x')) ||
							!isdigit(xptr[1])) {
						fprintf(stderr, "-s must be followed by WxH\n");
						return 1;
					}
					win_width = atoi(argv[i]);
					win_height = atoi(xptr + 1);
				}
				break;

			default:
				fprintf(stderr, "invalid argument: %s\n", argv[i]);
			case 'h':
				printf("Usage: xbench [options]\n");
				printf("Options:\n");
				printf(" -f      run colorfill test (XFillRectangle)\n");
				printf(" -b      run pixmap blit test (XCopyArea)\n");
				printf(" -s WxH  set the window dimensions\n");
				return argv[i][1] == 'h' ? 0 : 1;
			}
		} else {
			fprintf(stderr, "unexpected argument: %s\n", argv[i]);
			return 1;
		}
	}

	if(!(dpy = XOpenDisplay(0))) {
		fputs("failed to connect to the X server\n", stderr);
		return 1;
	}
	root_win = DefaultRootWindow(dpy);
	wm_del_win = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
	wm_prot = XInternAtom(dpy, "WM_PROTOCOLS", False);

	XpmCreatePixmapFromData(dpy, root_win, gnu_xpm, &pixmap, 0, 0);
	XGetGeometry(dpy, pixmap, &root_win, &tmp, &tmp,
			(unsigned int*)&pmap_width, (unsigned int*)&pmap_height,
			(unsigned int*)&tmp, (unsigned int*)&tmp);

	win = create_window(win_width, win_height, "Simple X benchmark");
	gc = XCreateGC(dpy, win, 0, 0);

	XMapWindow(dpy, win);
	XFlush(dpy);

	gettimeofday(&tv0, 0);

	do {
		XNextEvent(dpy, &event);
	} while(handle_event(&event));

	XSync(dpy, True);

	gettimeofday(&tv, 0);

	sec = tv.tv_sec - tv0.tv_sec + SEC(tv.tv_usec) - SEC(tv0.tv_usec);

	printf("frames: %d, seconds: %f\n", frames, sec);
	printf("fps: %f\n", frames / sec);

	XFreeGC(dpy, gc);
	XDestroyWindow(dpy, win);
	XCloseDisplay(dpy);
	return 0;
}

void redraw(void)
{
	static int x, y;

	if(test == TEST_FILL) {
		XSetForeground(dpy, gc, WhitePixel(dpy, DefaultScreen(dpy)));
		XFillRectangle(dpy, win, gc, x, y, pmap_width, pmap_height);
	} else {
		XCopyArea(dpy, pixmap, win, gc, 0, 0, pmap_width, pmap_height, x, y);
	}
	XFlush(dpy);

	frames++;

	if(++x >= win_width - pmap_width) {
		x = 0;
		if(++y >= win_height - pmap_height) {
			y = 0;
		}
	}

	post_redisplay();
}

Window create_window(int xsz, int ysz, const char *title)
{
	Window win, rwin;
	int scr;
	XSetWindowAttributes xattr;
	XTextProperty wmname;
	XClassHint chint;

	scr = DefaultScreen(dpy);
	rwin = RootWindow(dpy, scr);

	xattr.background_pixel = xattr.border_pixel = BlackPixel(dpy, scr);
	xattr.colormap = DefaultColormap(dpy, scr);

	win = XCreateWindow(dpy, rwin, 0, 0, xsz, ysz, 0, CopyFromParent, InputOutput,
			DefaultVisual(dpy, scr), CWColormap | CWBackPixel | CWBorderPixel, &xattr);
	XSelectInput(dpy, win, ExposureMask | StructureNotifyMask | KeyPressMask);

	XSetWMProtocols(dpy, win, &wm_del_win, 1);

	XStringListToTextProperty((char**)&title, 1, &wmname);
	XSetWMName(dpy, win, &wmname);
	XFree(wmname.value);

	chint.res_name = "xbench";
	chint.res_class = "xbench";
	XSetClassHint(dpy, win, &chint);

	return win;
}

int handle_event(XEvent *ev)
{
	KeySym keysym;
	static int mapped;

	switch(ev->type) {
	case MapNotify:
		mapped = 1;
		break;

	case UnmapNotify:
		mapped = 0;
		break;

	case Expose:
		if(mapped && ev->xexpose.count == 0) {
			redraw();
		}
		break;

	case KeyPress:
		keysym = XLookupKeysym((XKeyEvent*)&ev->xkey, 0);
		if((keysym & 0xff) == 27) {
			return 0;
		}
		break;

	case ClientMessage:
		if(ev->xclient.message_type == wm_prot
				&& ev->xclient.data.l[0] == wm_del_win) {
			return 0;
		}
		break;
	}

	return 1;
}

void post_redisplay(void)
{
	XEvent event;
	event.type = Expose;
	event.xexpose.type = Expose;
	event.xexpose.display = dpy;
	event.xexpose.window = win;
	event.xexpose.x = event.xexpose.y = 0;
	event.xexpose.width = win_width;
	event.xexpose.height = win_height;
	event.xexpose.count = 0;

	XSendEvent(dpy, win, False, 0, &event);
}
